#!/usr/bin/env python

# Copyright (C) 2002 Jrg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os.path
import random
import string
import re
import time

import mp3info

import events

from helper import formattime, debug

tracknrandtitlere = re.compile("^\[?(\d+)\]?-(.*)\.mp3$")
UNKNOWN = "Unknown"

def capword(word):
    for i in range(len(word)):
        if word[i] not in string.punctuation: break
    else:
        i = len(word)
    return word[:i+1].upper() + word[i+1:]

def capitalize(s):
    """capitalize first character of word, leave rest intact"""
    return string.join([capword(word) for word in s.split()])

class item:
    pass


class song(item):
    def __init__(self, path):
        assert os.path.isfile(path), "'%s' doesn't exist" % path
        self.path = os.path.normpath(path)
        self.name = os.path.basename(path)
        
        match = re.match(tracknrandtitlere, self.name)
        if match:
            fntracknr = str(int(match.group(1)))
            fntitle = match.group(2)
        else:
            fntracknr = ""
            fntitle = self.name
	    if fntitle.endswith(".mp3"):
	        fntitle = fntitle[:-4]

        try:
            id3info = mp3info.get_mp3tag(self.path)
            self.title = capitalize(id3info.get('TITLE', fntitle))
            self.album = capitalize(id3info.get('ALBUM', UNKNOWN))
            self.artist = capitalize(id3info.get('ARTIST', UNKNOWN))
            self.year = id3info.get('YEAR', "")
            self.genre  = id3info.get('GENRE', "")
            self.tracknr = id3info.get('TRACKNUMBER', fntracknr)
        except:
            self.title = fntitle
            self.album = UNKNOWN
            self.artist = UNKNOWN
            self.year = ""
            self.genre = ""
            self.tracknr = fntracknr

        # do some further treatment of the song info

        # use title from filename, if this is only a longer version of
        # the id3 tag title
        if fntitle.startswith(self.title):
	    self.title = fntitle

        # strip leading "The " in artist names, often used inconsistently
        if self.artist.startswith("The "):
            self.artist = self.artist[4:]

        # we don't want empty album names
        if self.album=="":
            self.album = UNKNOWN

        # nor empty artist names
        if self.artist=="":
            self.artist = UNKNOWN
            
        # nor empty artist names
        if self.genre=="":
            self.genre = UNKNOWN


        # get the length of the MP3 file
        try:
            mp3i = mp3info.mp3info(self.path)
            self.length = (mp3i["MM"]*60+mp3i["SS"])*1000
        except:
            self.length = 0

        self.nrplayed = 0
        self.lastplayed = None

    def __cmp__(self, other):
        return cmp(self.path, other.path)

    def __repr__(self):
        return self.path

    def play(self):
        self.nrplayed += 1
        self.lastplayed = time.time()
        events.hub.notify(events.songupdate(self))


class diritem(item):
    
    """ item containing other items """

    def cmpitems(self, x, y):
        """ compare to items in contained in self """
        pass
    
    def getcontents(self, songdb):
        """ return items contained in self (using database songdb) """
        pass

    def getheader(self, item, songdb):
        """ return header (used for title bar in filelistwin) of item in self.
        The songdb is also passed
        
        Note that item can be None!
        """
        pass


class artist(diritem):
    def __init__(self, name, path=None):
        self.name = name
        self.albums = {}
        self.songs = []
        self.genres = []
        self.path = path

    def cmpitem(self, x, y):
        return cmp(x.name, y.name)

    def getcontents(self, songdb):
        albums = songdb.getalbums(self.name)
        albums.sort(self.cmpitem)
        return albums

    def getheader(self, item, songdb):
        return self.name


class album(diritem):
    def __init__(self, artist, name, path=None):
        self.name = name
        self.artist = artist
        self.songs = []
        self.genres = []
        self.path = path

    def cmpitem(self, x, y):
        return ( x.tracknr!="" and y.tracknr!="" and
                 cmp(int(x.tracknr), int(y.tracknr)) or
                 cmp(x.name, y.name) or 
                 cmp(x.path, y.path) )

    def getcontents(self, songdb):
        songs = songdb.getsongs(self.artist, self.name)
        songs.sort(self.cmpitem)
        return songs

    def getheader(self, item, songdb):
        return self.artist + " - " + self.name

class playlist(diritem):
    def __init__(self, path):
        assert os.path.isfile(path), "'%s' doesn't exist" % path
        self.path = os.path.normpath(path)
        self.name = os.path.basename(path)
        if self.name.endswith(".m3u"):
            self.name = self.name[:-4]
        self.songs = []

        file = open(self.path, "r")
        
        for line in file.xreadlines():
            # XXX: interpret extended m3u format (especially for streams)
            # see: http://forums.winamp.com/showthread.php?s=dbec47f3a05d10a3a77959f17926d39c&threadid=65772
            if not line.startswith("#"):
                path = line.strip()
                if not path.startswith("/"):
                    path = os.path.join(self.path, path)
                if os.path.isfile(path):
                    self.songs.append(path)
        file.close()

    def cmpitem(self, x, y):
        # this method is not very well defined for a playlist,
        # since the order of items is defined by the playlist
        # itself, but just for completeness, we set:
        return cmp(x.name, y.name)

    def getcontents(self, songdb):
        return map(songdb.getsong, self.songs)

    def getheader(self, item, songdb):
        if item is not None:
            return item.artist + " - " + item.album
        else:
            return self.name

#
# virtual items, i.e. items not present in database
#

class filtereditem(diritem):

    """ item, which consists of filtered items """
    
    def __init__(self, fname, item, filterfunc):
        self.fname = fname
        self.item = item
        self.filterfunc = filterfunc
        self.name = "%s <%s>" % (self.item.name, self.fname)

    def cmpitem(self, x, y):
        return self.item.cmpitem(x, y)

    def getcontents(self, songdb):
        contents = []
        for item in self.item.getcontents(songdb):
            if self.filterfunc(item):
                if isinstance(item, song):
                    contents.append(item)
                else:
                    contents.append(filtereditem(self.fname, item, self.filterfunc))
        return contents

    def getheader(self, item, songdb):
        return "%s <%s>" % (self.item.getheader(item, songdb),
                            self.fname)


class genre(diritem):
    def __init__(self, name):
        self.name = name

    def cmpitem(self, x, y):
        return cmp(x.name, y.name)

    def filterfunc(self, item):
        try:
            return self.name in item.genres
        except AttributeError:
            return self.name==item.genre

    def getcontents(self, songdb):
        filteredartists = []
        for artist in songdb.getartists():
            if self.name in artist.genres:
                filteredartists.append(filtereditem("%s=%s" % (_("Genre"),self.name),
                                                    artist,
                                                    self.filterfunc))
                
        filteredartists.sort(self.cmpitem)
        return filteredartists

    def getheader(self, item, songdb):
        return self.name


class topplayed(diritem):
    def __init__(self, maxnr):
        self.name = "[%s]" % _("Top played songs")
        self.maxnr = maxnr

    def cmpitem(self, x, y):
        return cmp(y.nrplayed, x.nrplayed) or cmp(y.lastplayed, x.lastplayed)

    def getcontents(self, songdb):
        songs = songdb.getsongs()
        songs.sort(self.cmpitem)
        songs = [ song for song in songs[:self.maxnr]
                       if song.lastplayed is not None ]

        return songs

    def getheader(self, item, songdb):
        if item is not None:
            return item.artist + " - " + item.album
        else:
            return self.name


class randomlist(diritem):
    def __init__(self, maxnr):
        self.name = "[%s]" % _("Random song list")
        self.maxnr = maxnr

    def cmpitem(self, x, y):
        # just for completeness, we set:
        return cmp(x.name, y.name)

    def getcontents(self, songdb):
        songlist = songdb.getsongs()
        songs = []

        if len(songlist)!=0:
            for i in range(self.maxnr):
                songs.append(random.choice(songlist))
        
        return songs

    def getheader(self, item, songdb):
        if item is not None:
            return item.artist + " - " + item.album
        else:
            return self.name


class lastplayed(diritem):
    def __init__(self, maxnr):
        self.name = "[%s]" % _("Last played songs")
        self.maxnr = maxnr

    def cmpitem(self, x, y):
        return cmp(y.lastplayed, x.lastplayed)

    def getcontents(self, songdb):
        songs = songdb.getsongs()
        songs.sort(self.cmpitem)
        songs = [ song for song in songs[:self.maxnr]
                       if song.lastplayed is not None ]
        return songs

    def getheader(self, item, songdb):
        if item is not None:
            return item.artist + " - " + item.album
        else:
            return self.name


class genres(diritem):
    def __init__(self):
        self.name = "[%s]" % _("Genres")

    def cmpitem(self, x, y):
        return cmp(x.name, y.name)

    def getcontents(self, songdb):
        genres = songdb.getgenres()
        genres.sort(self.cmpitem)
        return genres

    def getheader(self, item, songdb):
        return _("Genres")


class playlists(diritem):
    def __init__(self):
        self.name = "[%s]" % _("Playlists")

    def cmpitem(self, x, y):
        return cmp(x.name, y.name)
        
    def getcontents(self, songdb):
        playlists = songdb.getplaylists()
        playlists.sort(self.cmpitem)
        return playlists

    def getheader(self, item, songdb):
        return _("Playlists")


class basedir(diritem):
    """ base dir of database view"""
    
    def __init__(self, mp3basedir, maxnr):
        self.name = _("MP3 Database")
        self.mp3basedir = mp3basedir
        self.maxnr = maxnr
        self.filesystembase = filesystemdir(self.mp3basedir, self.mp3basedir)
        self.topplayed = topplayed(self.maxnr)
        self.genres = genres()
        self.lastplayed = lastplayed(self.maxnr)
        self.randomlist = randomlist(self.maxnr)
        self.playlists = playlists()

    def cmpitem(self, x, y):
        return cmp(x.name, y.name)
        
    def getcontents(self, songdb):
        contents = songdb.getartists()
        contents.sort(self.cmpitem)
        contents.append(self.filesystembase)
        contents.append(self.genres)
        contents.append(self.topplayed)
        contents.append(self.lastplayed)
        contents.append(self.randomlist)
        contents.append(self.playlists)
        return contents

    def getheader(self, item, songdb):
        return "%s (%d %s)" % ( _("MP3 Database"),
                                len(songdb.songs),
                                _("entries"))


class filesystemdir(diritem):

    """ diritem corresponding to directory in filesystem """

    def __init__(self, mp3basedir, dir):
        self.mp3basedir = mp3basedir
        self.dir = dir
        
        if self.dir==self.mp3basedir:
            self.name = "[%s]" % _("Filesystem")
        else:
            self.name = self.dir[len(self.mp3basedir):].split("/")[-1]

    def cmpitem(self, x, y):
        return cmp(x.name, y.name)

    def getcontents(self, songdb):
        items = []
        try:
            for name in os.listdir(self.dir):
                try:
                    path = os.path.join(self.dir, name)
                    if os.path.isdir(path) and os.access(path, os.R_OK|os.X_OK):
                        newitem = filesystemdir(self.mp3basedir, path)
                        items.append(newitem)
                    elif name.endswith(".mp3") and os.access(path, os.R_OK):
                        newsong = songdb.queryregistersong(path)
                        items.append(newsong)
                except (IOError, OSError) : pass
        except OSError:
            return None
        return items

    def getheader(self, item, songdb):
        if self.dir==self.mp3basedir:
            return _("Filesystem")
        else:
            return self.name


    
