#!/usr/bin/env python

# Copyright (C) 2002 Jrg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import shelve
import threading

import item
import events

from helper import debug

#
# ReadWriteLock from Python Cookbook Recipe 6.3
#

class ReadWriteLock:
    def __init__(self):
        self._read_ready = threading.Condition(threading.Lock())
        self._readers = 0

    def acquire_read(self):
        # debug("ReadWriteLock: acquire read\n")
        self._read_ready.acquire()
        try:
            self._readers += 1
            # debug("ReadWriteLock: %d readers\n" % self._readers)
        finally:
            self._read_ready.release()

    def release_read(self):
        # debug("ReadWriteLock: release read\n")
        self._read_ready.acquire()
        try:
            self._readers -= 1
            # debug("ReadWriteLock: %d readers\n" % self._readers)
            if not self._readers:
                self._read_ready.notifyAll()
        finally:
            self._read_ready.release()

    def acquire_write(self):
        # debug("ReadWriteLock: acquire write\n")
        self._read_ready.acquire()
        while self._readers>0:
            self._read_ready.wait()

    def release_write(self):
        # debug("ReadWriteLock: release write\n")
        self._read_ready.release()

#
# songdb class
#

class songdb:
    def __init__(self, channel, songdbbase=None):
        self.songdbbase = songdbbase
        
        if self.songdbbase:
            self.songs = shelve.open(self.songdbbase + "_songs.db")
            self.artists = shelve.open(self.songdbbase + "_artists.db")
            self.playlists = shelve.open(self.songdbbase + "_playlists.db")
            
            debug("songdb: %d songs, %d artists and %d playlists registered\n" %
                  (len(self.songs), len(self.artists), len(self.playlists)))
        else:
            self.songs = {}
            self.artists = {}
            self.playlists = {}

        # empty cache of all artists and all songs
        # This variables contain tuples acting as caches for
        # all artists and all songs, respectively, in the database
        self.allartists = self.allsongs = None

        # this lock protects write access to the database
        self.rwlock = ReadWriteLock()

        # we need to be informed about song info changes
        channel.subscribe(events.songupdate, self.songupdate)

    def close(self):
        if self.songdbbase:
            self.songs.close()
            self.artists.close()
            self.playlists.close()

    def queryregistersong(self, path):
        """get song info from database or insert new one"""

        self.rwlock.acquire_write()
        
        if self.songs.has_key(path):
            song = self.songs[path]
        else:
            debug("songdb: new song: %s\n" % path)
            song = self.songs[path] = item.song(path)
            self.allsongs = None

        # insert song in artist index, if neccesary
        if not self.artists.has_key(song.artist):
            debug("songdb: new artist: %s\n" % song.artist)
            newartist = item.artist(song.artist)
            self.artists[song.artist] = newartist
            self.allartists = None
            events.hub.notify(events.artistaddedtosongdb(newartist))

        # now, we update the artist information
        artist = self.artists[song.artist]
        
        if not self.artists[song.artist].albums.has_key(song.album):
            newalbum = item.album(song.artist, song.album)
            artist.albums[song.album] = newalbum
            events.hub.notify(events.albumaddedtosongdb(newalbum))

        if path not in artist.albums[song.album].songs:
            artist.albums[song.album].songs.append(path)
            # note that for performance reasons, we don't send notifications
            # of song additions

        if song.genre not in artist.albums[song.album].genres:
            artist.albums[song.album].genres.append(song.genre)

        if song.genre not in artist.genres:
            artist.genres.append(song.genre)

        if artist!=self.artists[song.artist]:
            self.artists[song.artist] = artist

        self.rwlock.release_write()
            
        return song

    def delsong(self, path):
        """delete song from database"""

        # XXX: not very well tested

        self.rwlock.acquire_write()
        
        if not self.songs.has_key(path):
            raise "Cannot delete nonexistent song"

        debug("songdb: delete song: %s\n" % path)
        song = self.songs[path]

        # update artist and/or album info
        artist = self.artists[song.artist]
        album = artist.albums[song.album]
        
        artist.songs.remove(song.path)
        album.songs.remove(song.path)
        for asong in album.songs:
            if asong.genre==song.genre:
                break
        else:
            album.genres.remove(song.genre)
            for asong in artist.songs:
                if asong.genre==song.genre:
                    break
            else:
                artist.genres.remove(song.genre)
                
        self.artists[song.artist] = artist

        # delete album and artist, if necessary
        if len(album.songs)==0:
            del artist.albums[song.album]
            self.artists[song.artist] = artist

            if len(artist.albums)==0:
                del self.artists[song.artist]
                debug("songdb: deleted artist: %s\n" % song.artist)
                self.allartists = None
                # XXX: no event

        del self.songs[path]
        self.allsongs = None

        self.rwlock.release_write()

    def registerplaylist(self, path):
        playlist = item.playlist(path)

        # also try to register songs in playlist
        # delete song, if this fails
        for i in range(len(playlist.songs)):
            try:
                self.queryregistersong(playlist.songs[i])
            except (IOError, OSError):
                del playlist.songs[i]

        # a resulting, non-empty playlist can be written in the database
        if playlist.songs:
            self.rwlock.acquire_write()
            self.playlists[path] = playlist
            self.rwlock.release_write()

    def updatesong(self, song):
        """updates entry of given song"""
        # XXX: changes of artist, album and genre information are not handled
        self.rwlock.acquire_write()

        # update song cache if existent
        try:
            # Note that we rely on each song being uniquely specified
            # by its path and the corresponding __cmp__ method of
            # item.song!
            oldsong = self.songs[song.path]
            self.allsongs[self.allsongs.index(oldsong)] = song
        except (ValueError, AttributeError):
            pass

        self.songs[song.path] = song
        
        self.rwlock.release_write()
        
    def getsong(self, path):
        """returns song entry with given path"""
        self.rwlock.acquire_read()
        song = self.songs.get(path)
        self.rwlock.release_read()
        return song

    def getartists(self):
        """return all stored artists"""
        self.rwlock.acquire_read()
        # use cached value if existent
        if self.allartists:
            artists = self.allartists[:]
            self.rwlock.release_read()
            return artists
        else:
            artists = map(self.artists.get, self.artists.keys())
            self.rwlock.release_read()
            self.rwlock.acquire_write()
            self.allartists = artists[:]
            self.rwlock.release_write()
            return artists

    def getalbums(self, artist=None):
        """return albums of a given artist.

        artist has to be a string. If it is none, all stored
        albums are returned
        """
        if artist is None:
            albums = []
            for artist in self.getartists():
                for album in self.getalbums(artist.name):
                    albums.append(album)
            return albums
        else:
            self.rwlock.acquire_read()
            albums = self.artists[artist].albums.values()
            self.rwlock.release_read()
            return albums

    def getsongs(self, artist=None, album=None):
        """ returns song of given artist and given album
        
        artist and album are either strings or None
        If they are not given all artists and/or albums are chosen
        """
        
        if artist is None and album is None:
            # return all songs in songdb
            self.rwlock.acquire_read()
            # use cached value if existent
            if self.allsongs:
                songs = self.allsongs[:]
                self.rwlock.release_read()
                return songs
            else:
                songs = map(self.getsong, self.songs.keys())
                self.rwlock.release_read()
                self.rwlock.acquire_write()
                self.allsongs = songs[:]
                self.rwlock.release_write()
                return songs
        elif album is None:
            # return all songs of a given artist
            self.rwlock.acquire_read()
            songs = map(self.getsong, self.artists[artist].songs)
            self.rwlock.release_read()
            return songs
        elif artist is None:
            # return all songs which are on albums with the given name
            return [ aalbum for aalbum in self.getalbums()
                            if aalbum.name==album ]
        else:
            # return all songs on a given album of a given artist
            self.rwlock.acquire_read()
            songs = map(self.getsong,
                       self.artists[artist].albums[album].songs)
            self.rwlock.release_read()
            return songs

    def getgenres(self):
        genres = []
        for song in self.getsongs():
            if song.genre and song.genre not in genres:
                genres.append(song.genre)
        genres = map(item.genre, genres)
        return genres

    def getplaylist(self, path):
        """returns playlist entry with given path"""
        self.rwlock.acquire_read()
        playlist = self.playlists.get(path)
        self.rwlock.release_read()
        return playlist

    def getplaylists(self):
        self.rwlock.acquire_read()
        # use cached value if existent
        
        playlists = map(self.getplaylist, self.playlists.keys())
        self.rwlock.release_read()
        return playlists

    def registerdirtree(self, dir):
        for name in os.listdir(dir):
            path = os.path.join(dir, name)
            if os.access(path, os.R_OK):
                if os.path.isdir(path):
                    try: self.registerdirtree(path)
                    except (IOError, OSError): pass
                elif name.endswith(".mp3"):
                    try: self.queryregistersong(path)
                    except (IOError, OSError): pass
                elif name.endswith(".m3u"):
                    try: self.registerplaylist(path)
                    except (IOError, OSError): pass

    # event handler

    def songupdate(self, event):
        self.updatesong(event.song)

#
# thread for automatic registering of songs in database
#

class songautoregisterer(threading.Thread):
    def __init__(self, db, basedir):
        self.db = db
        self.basedir = basedir

        threading.Thread.__init__(self)
        self.setDaemon(1)

    def run(self):
        self.db.registerdirtree(self.basedir)

        # fill up caches
        self.db.getsongs()
        self.db.getartists()
        debug("songautoregisterer: exit\n")
